# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Common utilities for BiDi command construction."""

from __future__ import annotations

from collections.abc import Generator
from typing import Any


def command_builder(
    method: str, params: dict[str, Any] | None = None
) -> Generator[dict[str, Any], Any, Any]:
    """Build a BiDi command generator.

    Args:
        method: The BiDi method name (e.g., "session.status", "browser.close")
        params: The parameters for the command. If omitted, an empty
            dictionary is sent.

    Yields:
        A dictionary representing the BiDi command

    Returns:
        The result from the BiDi command execution
    """
    if params is None:
        params = {}
    result = yield {"method": method, "params": params}
    return result
